from abc import ABC
from typing import Optional


class Headers:
    def __init__(self, role=None, user_id=None, key=None, **kwargs):
        headers = {"x-gapo-role": role or "service"}

        if key:
            headers.update({"x-gapo-api-key": key})

        if user_id:
            headers.update({"x-gapo-user-id": user_id})

        if kwargs:
            headers.update(kwargs)

        self._h = headers

    def __repr__(self):
        return str(self._h)

    def dict(self):
        return self._h

    def update(self, **custom_headers):
        if custom_headers.get("user_id"):
            user_id = custom_headers["user_id"]
            self._h.update({"x-gapo-user-id": user_id})
            del custom_headers["user_id"]

        if custom_headers.get("role"):
            role = custom_headers["role"]
            self._h.update({"x-gapo-role": role})
            del custom_headers["role"]

        if custom_headers.get("key"):
            key = custom_headers["key"]
            self._h.update({"x-gapo-api-key": key})
            del custom_headers["key"]

        if custom_headers:
            self._h.update(custom_headers)


class AbstractRequest(ABC):
    service: str
    endpoint: str = ""
    key: str
    response: Optional
    Response: Optional
    Params: Optional

    def __init__(self, client_cfg: dict, client_session):
        self.url = client_cfg[self.service] + "/" + self.endpoint

        api_key = client_cfg.get(
            getattr(self, "key", None) or (self.service + "___KEY")
        )

        self.headers = Headers(
            role=getattr(self, "role", None),
            key=api_key,
            user_id=getattr(self, "user_id", None),
        )

        self._cl = client_session



""" Application config template
"""
from os import environ
from typing import TypeVar, Union, Dict
from configparser import ConfigParser

ValidatorType = TypeVar("T")


def base_load_config(
    config_path: str,
    validator_cls: ValidatorType = None,
) -> Union[Dict, ValidatorType]:
    """Load and produce AppConfig, setting log level as well
    - Accept INI format only
    - Final value is ALWAYS UPPERCASE
    """
    config = {}
    stage = environ.get("STAGE", "DEVELOPMENT").upper()
    parser = ConfigParser()
    parser.read(config_path)

    for k, val in parser.items(stage):
        key = k.upper()
        value = val or environ.get(key)
        config.update({key: value})

    if validator_cls:
        config = validator_cls(**config)

    return config


def load_config_as_dict(config_path: str = "config.ini") -> Dict:
    """load config and return a dict"""
    return base_load_config(config_path)


def load_config_as_class(
    validator_cls: ValidatorType,
    config_path: str = "config.ini",
) -> ValidatorType:
    """load config and return a class instance,
    maybe Dataclass or a Pydantic Model
    """
    return base_load_config(config_path, validator_cls)



""" Least-Recently-Used In-Process Caching
"""
from typing import Tuple, Optional
from gapoic.abs import DoublyLinkedList, Node
from gapoic.exceptions import ReadonlyException


class LRU:
    """LRU that accepts flexible data"""

    def __init__(self, name: str = "", *, limit: int = 0):
        if not isinstance(limit, int):
            raise ValueError("Invalid lru max-size")

        if limit < 1:
            raise ValueError("Invalid lru max-size")

        self.__limit = limit
        self.__ll = DoublyLinkedList()
        self.__map = dict()
        self.__name = name
        self.__tail = None

    def __str__(self):
        ll_as_str = str(self.__ll)
        return ll_as_str

    def __getitem__(self, key) -> Optional[Node]:
        """short-syntax for access node in lru"""
        return self.get(key)

    def get(self, key) -> Optional[Node]:
        """Get a node from lru"""
        node = self.__map.get(key)

        if not node:
            return None

        # NOTE: move node to top:
        if self.__ll.head is not node:
            node.prev.next = node.next
            if node.next:
                node.next.prev = node.prev
            else:
                self.__tail = node.prev
            node.next = self.__ll.head
            self.__ll.head.prev = node
            self.__ll.head = node
            node.prev = None

        return node

    def put(self, key, data) -> Optional[Node]:
        """Putting data in LRU
        - if existing data, move it to top of the list
        - if non-exist, create new node if limit is
        not reached and insert it at the head of the list
        - return the evicted node if neccessary
        """
        node = self.__map.get(key)

        if not node:
            node = Node(key, data, mutable=True)
            self.__map.update({key: node})

            if self.__ll.head:
                node.next = self.__ll.head
                self.__ll.head.prev = node

            self.__ll.head = node
        else:
            node.data = data

            if self.__ll.head is not node:
                node.prev.next = node.next
                if node.next:
                    node.next.prev = node.prev
                else:
                    self.__tail = node.prev
                node.next = self.__ll.head
                self.__ll.head.prev = node
                self.__ll.head = node
                node.prev = None

        if not self.__tail:
            self.__tail = self.__ll.head

        if self.size > self.limit:
            tail = self.__tail
            self.__tail = tail.prev
            self.__tail.next = None

            del self.__map[tail.key]
            return tail
        return None

    @property
    def limit(self):
        """Maximum number of node that the LRU can hold"""
        return self.__limit

    @limit.setter
    def limit(self, _):
        """read-only"""
        raise ReadonlyException("LRU's limit")

    @property
    def size(self):
        """Current number of nodes that the LRU is keeping"""
        return len(self.__map)

    @size.setter
    def size(self, _):
        """read-only"""
        raise ReadonlyException("LRU's size")

    def link_ends(self) -> Tuple[Optional[Node], Optional[Node]]:
        """Used for debugging mostly
        return an immutable copy of head & tail of the lru
        """
        if not self.__ll.head:
            return None, None

        copied_head = Node(self.__ll.head.key, self.__ll.head.data)
        copied_tail = Node(self.__tail.key, self.__tail.data)
        return copied_head, copied_tail



"""Various helping functions"""
from typing import Callable, Iterable
from time import time
from functools import wraps

try:
    from logzero import logger
except ImportError:
    import logging as logger

# =======================================================================
# Class Utitlities
# =======================================================================


class Singleton(type):
    """Singleton Metaclass"""

    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(
                *args,
                **kwargs,
            )
        return cls._instances[cls]


# =======================================================================
# Function utitlities
# =======================================================================
def now(
    secs: int = 0,
    mins: int = 0,
    hrs: int = 0,
    days: int = 0,
    milisecond: bool = False,
):
    """
    Get time in Unix
    Args:
        sec : Seconds
        min : Minutes
        hr  : Hours
        day : Days
        milisecond : miliseconds or seconds
    Returns:
        int : unix time
    """
    current = int(time())
    delta = secs * 1 + mins * 60 + hrs * 3600 + days * 3600 * 24
    result = current + delta
    return result * 1000 if milisecond else result


def find(predicate: Callable, items: Iterable, default=None):
    """Find first item in the iterable
    that matches using predicate function
    - A predicate function must always return a boolean value
    """
    return next((x for x in items if predicate(x) is True), default)


def try_except(description: str, propagate=False):
    """Wrapping function in try-except block, allow propagate to upper level"""

    def wrapped(func):
        @wraps(func)
        def inner_wrapped(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as err:
                msg = f"=====>>> {description} <<<======"
                logger.error(msg)
                logger.exception(err)
                logger.error("Args: %s", args)
                if propagate:
                    raise err

        return inner_wrapped

    return wrapped



from pydantic import BaseModel, ValidationError
from .base import AbstractRequest

try:
    from logzero import logger
except ImportError:
    import logging as logger


class Request(AbstractRequest):
    def get(self, **params):
        try:
            if hasattr(self, "Params"):
                params = self.Params(**params).dict()

            headers = self.headers.dict()
            request_cfg = {"params": params, "headers": headers}
            resp = self._cl.get(self.url, **request_cfg)

            if resp.status_code != 200:
                msg = "Url=%s, API call unsuccesful => %s"
                text = resp.text
                logger.error(msg, self.url, text)
                return self.on_failure(resp)

            json = resp.json()
            return self.on_success(json)

        except ValidationError as err:
            logger.error("Url=%s, Invalid query-param: %s", self.url, err)
            return self.on_error(err)

        except Exception as err:
            logger.error("Url=%s, API call error: %s", self.url, err)
            return self.on_error(err)

    def on_failure(self, response):
        pass

    def on_error(self, error):
        pass

    def on_success(self, json_data):
        if hasattr(self, "Response"):
            return self.Response(**json_data)

        if hasattr(self, "response"):
            return self.response(**json_data)

        return json_data
